/*
 * AndroidPortAdditions.h
 *
 *  Created on: Jan 14, 2013
 *      Author: omergilad
 */

#ifndef ANDROIDPORTADDITIONS_H_
#define ANDROIDPORTADDITIONS_H_

#include "common/rect.h"
#include "common/hash-str.h"
#include "graphics/surface.h"
#include "graphics/scaler/hq.h"

#include "engines/engine.h"
#include "common/error.h"
#include "common/mutex.h"

#include "AndroidPortGlTextureHelper.h"
#include "AndroidPortUtils.h"
#include "GameAreaIdentifier.h"
#include "HitAreaHelper.h"
#include "graphics/Drawable.h"
#include "graphics/AlphaAnimation.h"

#include "subtitles/SubtitlePlayer.h"

#include <GLES/gl.h>

#include <list>
#include <queue>

#include "AndroidBitmap.h"
#include "CursorAndroidBitmap.h"

using Graphics::Surface;
using Graphics::PixelFormat;

using namespace Common;

/**
 * NONE - no touch going on
 * DOWN - touch is currently on (scroll, tap or hold)
 * UP - finger was just raised
 */
enum TouchState {
	NONE = 0, DOWN, UP, TAP, DOUBLE_TAP, FLING
};

struct TouchEvent
{
	TouchState mTouchState;
	uint16 mTouchX;
	uint16 mTouchY;
};

typedef Common::HashMap<String, AndroidBitmap*, IgnoreCase_Hash,
		IgnoreCase_EqualTo> BitmapHashMap;

class AndroidPortAdditions {

public:

	static inline AndroidPortAdditions* instance() {
		if (sInstance == NULL) {
			sInstance = new AndroidPortAdditions();
		}

		return sInstance;
	}

	static void release();

	bool shouldUseModifiedGamePixels();

	void setClassicMode(bool touchpadMode);

	bool getClassicMode();

	/**
	 * Translates the received touch coordinates if the game screen was adjusted.
	 */
	void translateTouchCoordinates(int& x, int& y);

	byte* getModifiedGamePixels();

	void onShowMouse(bool visible);

	void setDisplayDimensions(uint32 width, uint32 height);

//	Rect getDirtyRect();

	void initGLESResources();

	inline ShaderProgram* getDefaultShaderProgram() {
		return mDefaultShaderProgram;
	}

	inline ShaderProgram* getBlackShaderProgram() {
			return mBlackShaderProgram;
		}

	inline ShaderProgram* getScalerShaderProgram() {
		return mScalerShaderProgram;
	}

	void setGameTextureInfo(GLuint gameTexture, uint16 texWidth,
			uint16 texHeight);

	void onDrawTextureToScreen(GLshort x, GLshort y, GLshort w, GLshort h);

	void addBitmapResource(String key, AndroidBitmap* bitmap);

	void onMouseClick(int x, int y);

	bool onTapEvent(int x, int y, bool doubleTap = false);

	bool onLongClickEvent(int x, int y);

	bool onDoubleTapEvent(int x, int y);

	bool onScrollEvent(int x, int y);

	bool onFlingEvent(int x1, int y1, int x2, int y2);

	bool onDownEvent(int x, int y);

	bool onUpEvent(int x, int y);

	void beforeDrawTextureToScreen(Surface* gameSurface);

	bool saveGame(int32 slot);

	bool loadGame(int32 slot);

	void setSlotToLoad(int32 slot, bool force = false);

	void setSlotToSave(int32 slot, bool force = false);

	void setAutoLoadSlot(int32 slot);

	/**
	 * Returns whether we need to skip movies immediately without a fade out animation - used for auto-load
	 */
	bool isInAutoloadState();

	uint16* scale(const uint8 *srcPtr, uint16 x, uint16 y, uint16 w, uint16 h);

	inline uint16 getScalingFactor() {

		return mScalingFactor;
	}

	inline float getLQShaderScalingFactor() {

		return mLQShaderScalingFactor;
	}

	void setScalingOption(uint16 option);

	inline uint16 getScalingOption() {
		return mScalingOption;
	}

	void addShaderSource(const char* source, uint32 type);

	void gameEvent(int32 type);

	inline bool shouldMeasureRenderTime() {
		return mShouldTestShader;
	}

	void onRenderTimeMeasure(uint64 time);

	/**
	 * Should be called from the engine when there's a 'tick' that counts toward the idle state (if there is)
	 */
	void onGameIdleCounter();

	/**
	 * Should be called from the engine when there's an action that cancels the idle state.
	 * Used for protecting game saves when they should not be done during animations.
	 */
	void onGameNotIdle();

	bool checkLoadConditions();

	void onFastFadeInStarted();

	void onSystemInitialized();

	void onEngineInitialized();

	bool shouldUseLongClickAction();

	/**
	 * Should be called by the engine when the chosen action has changed
	 */
	void onActionChanged(uint16 action);

	/**
	 * Should be called by the engine when an action was clicked
	 */
	void onActionClicked(uint16 action);

	inline uint16 getShaderScalingMaxResolutionW() {
		return mShaderScalingMaxResolutionW;
	}

	inline uint16 getShaderScalingMaxResolutionH() {
		return mShaderScalingMaxResolutionH;
	}

	inline void setAfterGameIntro(bool value) {
		LOGD("AndroidPortAdditions: setAfterGameIntro: ");
		mAfterGameIntro = value;
	}

	uint32 displayToGameCoordinatesX(uint displayX);
	uint32 displayToGameCoordinatesY(uint displayY);
	uint32 gameToDisplayCoordinatesX(uint gameX);
	uint32 gameToDisplayCoordinatesY(uint gameY);

	inline float getDrawTransformationWidth() {
		return mDrawTransformationWidth;
	}

	inline float getDrawTransformationHeight() {
		return mDrawTransformationHeight;
	}

	bool isSimonGame();

	uint16 getGameType();

	void enableOpenHouseMode();

	void onVideoPlaying(String videoId);

	/**
	 * Called each time we want to update the subtitle state - should be called per game loop iteration
	 */
	void videoSubtitleTick();

	inline std::string getGamePath()
	{
		return mGamePath;
	}

	inline void setGamePath(std::string path)
	{
		mGamePath = path;
	}

	void onGamePaused();

	void onGameResumed();

	bool isT7GMapShowing();

	void performSkip(bool showAnimation = true);

	void performRevealItems();

	inline void setUpperTextIsLong(bool val)
	{
		mUpperTextIsLong = val;
	}

	/**
	 * Useful step for doing actions during any time the game engine ticks (regardless of graphics)
	 */
	void onGameEngineTick();

private:

	bool checkGameOverlayClicks(int x, int y, bool performOverlayAction = true);

	bool checkClick(uint32 x, uint32 y, uint32 buttonX, uint32 buttonY,
			uint32 buttonW, uint32 buttonH);

	AndroidBitmap* getBitmap(String key);

	void drawBitmapAsGlTexture(AndroidBitmap* bitmap, float x, float y,
			float width = 0, float height = 0, float alpha = 1.0, ShaderProgram* customProgram = NULL);

	void drawAnimationDrawable(const DrawablePtr drawable);

	void drawAnimations();

	void drawClassicModeMousePointer();

	void drawHotspotIndicator(AndroidBitmap* bitmap, int x, int y,
			AndroidBitmap* action, float alpha = 1.0);

	void drawNewBottomPanelSimon2();

	void drawSelectionOverlayFOTAQ();

	void setupCurrentHotspotAnimation(Hotspot& hotspot);

	void clearCurrentHotspotAnimation(long fadeoutDuration,
			bool completeFadeIn, bool actionUsed);

	void clearCurrentActionIconAnimation(bool actionPerformed);

	/**
	 * hotspotDrawable, actionDrawable - output arguments
	 */
	void generateHotspotIndicatorDrawables(AndroidBitmap* bitmap, Hotspot& hotspot,
			AndroidBitmap* action, DrawablePtr hotspotDrawable,
			DrawablePtr actionDrawable, DrawablePtr hotspotRectDrawable, float alpha = 1.0);

	void adjustHotspotIndicatorForFinger(uint16 touchX, uint16 touchY);

	void addBigActionFadeAnimation(AndroidBitmap* bitmap);

	void addHotspotFadeoutAnimation(Point hotspot, long duration);

	void addActionBlinkAnimation(Point hotspot, AndroidBitmap* action);

	void addWalkFadeoutAnimation(int x, int y);

	void initGlTextures();

	void initBitmapInGlTextures(AndroidBitmap* bitmap, float width,
			bool sourceContainsAlpha = false);


	void toggleBottomPanel();

	void performInventoryScroll(bool up);

	void moveBlackPanel(Surface* gameSurface);

	void checkGameInChat(Surface* gameSurface);

	void checkGameInPostcard(Surface* gameSurface);

	void checkBottomToolbar(Surface* gameSurface);

	/**
	 * Copy a rectangle between 2 surfaces.
	 * Assumes same pixel format.
	 */
	void copyPixelsBetweenSurfaces(Surface* src, Surface* dst, Rect srcRect,
			Point dstPoint, bool startFromBottom = false);

	//void copyPixelsToMagnifier();

	void requestMagnifierForCoordinates(int32 x, int32 y);

	bool isValidMagnifierCoordinates(int32 x, int32 y);

	void pushClickEvent(int32 x, int32 y, bool rightClick = false);

	void pushScrollEvent(int x, int y);

	//Action recognizeActionPress(int32 x, int32 y);

	AndroidBitmap* getActionIcon(uint16 action);

	void preventIdleMode();

	void gameTouchBehavior();

	void bottomToolbarClickBehavior(int16 gameTouchX, int16 gameTouchY);

	bool canSkip();

	bool canReplay();

	bool canShowMap();

	bool canShowRevealItems();

	bool canShowMenuButton();

	/**
	 * Safety measure for touch events - makes sure the events fall into the known surface dimensions.
	 * Solves an issue on Jelly Bean where soft buttons trigger irrelevant touch events.
	 */
	inline void touchEventSafety(int& x, int& y) {
		if (x < 0)
			x = 0;
		if (y < 0)
			y = 0;
		if (x >= mDisplayWidth)
			x = mDisplayWidth - 1;
		if (y >= mDisplayHeight)
			y = mDisplayHeight - 1;
	}

	void autoloadBehaviorSimon();

	void loadIfNeeded();

	void saveIfNeeded();

	void saveProtectionSimon1();

	void saveProtectionSimon2();

	bool checkSaveConditions();

	void resetSaveState();

	void forceAutosave();

	/**
	 * Set the current game touched point according to screen touch
	 */
	void gameTouchEvent(int16 x, int16 y, int16 origX, int16 origY,
			TouchState touchState);

	void assistWalkCoordinates(int16& x, int16& y);

	void chatArrowClick(bool up);

	void chatButtonClick();

	uint16 getCurrentAction();

	void fallbackToSoftwareScaler();

	void fallbackToLQHardwareScaler();

	void outsideDragonCaveWorkaround(int16 &x, int16 &y);

	bool isOpenHouseGrid();

	inline float adjustToDrawTransformationX(float x) {
		return 0.5 - mDrawTransformationWidth / 2 + x * mDrawTransformationWidth;
	}
	inline float adjustToDrawTransformationY(float y) {
		return 0.5 - mDrawTransformationHeight / 2
				+ y * mDrawTransformationHeight;
	}

	uint16 mGameType;
	BitmapHashMap mBitmaps;
	GLuint* mGlTextures;

	AndroidPortGlTextureHelper mGlTextureHelper;

	bool mMouseVisible;
	bool mMouseJustTurnedInvisible;


	uint16 mDisplayWidth;
	uint16 mDisplayHeight;

	float mLowerButtonPosY;

// Width / height
	float mDisplayRatio;

	Surface mBlackPanelTempSurface;
	Surface mModifiedGameSurface;

	bool mShouldUseModifiedGamePixels;

	bool mClassicMode;
	bool mShouldShowNewActionBar;

	bool mGameInChat;
	bool mBottomToolbarAppearing;
	bool mGameInPostcard;

	GameArea mCurrentGameArea;

	uint16 mTouchTranslationOffset;
	uint16 mTouchTranslationLowerBound;

	bool mDumpPixelsToFile;

	std::queue<TouchEvent> mTouchEventQueue;
	Common::Mutex* mTouchEventMutex;

	Hotspot mCurrentHotspot;
	bool mDisallowNextTouchEvent;

	DrawablePtr mCurrentHotspotDrawable;
	DrawablePtr mCurrentHotspotRectDrawable;
	DrawablePtr mCurrentActionDrawable;

	bool mPreventFingerCoveringHotspot;

	uint16 mRevealItemsFadeCounter;

	float mGameToDisplayRatioW;
	float mGameToDisplayRatioH;

	uint16 mCurrentAction;

	uint16 mActionIconFadeCounter;
	uint16 mSkipIconFadeCounter;

	bool mGameDisplayStarted;

	uint64 mLastTouchMoveTimestamp;

	uint64 mFirstDrawTimestamp;

	int32 mSlotToLoad;
	int32 mSlotToSave;
	uint64 mSaveStartTimestamp;
	bool mSaveInProgress;
	bool mGameIdleCountWhileSaving;
	bool mGameCancelledIdleWhileSaving;
	bool mForceSaveLoad;

	int32 mAutoLoadSlot;
	uint mAutoLoadSpamSkipCounter;
	bool mAutoLoadState;
	uint64 mAutoLoadIntroEndTimestamp;

	bool mGameStartedInAutoload;

	uint16* mScaledOutput;

	ScalerPluginObject* mScalerPlugin;

	uint16 mScalingFactor;
	float mLQShaderScalingFactor;

	uint16 mScalingOption;

	ShaderProgram* mDefaultShaderProgram;
	ShaderProgram* mBlackShaderProgram;

	ShaderProgram* mScalerShaderProgram;
	ShaderProgram* mRectHighlightingShaderProgram;

	bool mShouldTestShader;
	uint16 mRenderMeasureSkippedFrameCount;
	uint16 mRenderMeasureCount;
	double mRenderMeasureAverage;
	uint64 mShaderTestStartTime;

	uint64 mLastIdlePreventionTimestamp;

	bool mShouldPerformRevealItems;

	uint64 mLastTimeSkipPressed;

	uint64 mLastTimeRevealItemsPressed;

	const char* mScalerVertexSource;
	const char* mScalerFragmentSource;
	const char* mScalerLQVertexSource;
	const char* mScalerLQFragmentSource;
	const char* mRectHighlightingVertexSource;
	const char* mRectHighlightingFragmentSource;

	uint16 mGameTextureW;
	uint16 mGameTextureH;
	GLuint mGameTexture;

// Screen transformation when maintaining aspect ratio
	float mDrawTransformationWidth, mDrawTransformationHeight;

	bool mSkipPressedAtLeastOnce;

	HitAreaHelper mHitAreaHelper;

	std::list<DrawablePtr> mAnimatedDrawables;

	AndroidBitmap* mCursorBitmap;

	// Bitmap that draws nothing, for use in shader effects
	AndroidBitmap* mDummyBitmap;

// 1-based index, 0 means no selection
	int16 mSelectedChatRow;

	bool mShouldPerformSimon1PuddleWorkaround;

	bool mStartInOpenHouse;

// Relevant for T7G, for knowing whether we're after the long game intro
	bool mAfterGameIntro;

// Maximum resolution for shader scaling (for performance reasons)
	uint16 mShaderScalingMaxResolutionW;
	uint16 mShaderScalingMaxResolutionH;

	uint16* mEmptyClearingBuffer;

	SubtitlePlayer mSubtitlePlayer;

	std::string mGamePath;

	bool mDebugHideIconsToggle;

	bool mUseFotaqHotspotWorkaround;
	bool mUpperTextIsLong;

	static AndroidPortAdditions* sInstance;

	AndroidPortAdditions();
	virtual ~AndroidPortAdditions();
};

#endif /* ANDROIDPORTADDITIONS_H_ */
